(function () {
    'use strict';

    const IDS = ['background', 'sidebar', 'accent'];

    Macro.add('theme', {
        handler: function () {
            let target, effect, value, variable = null, output;
            if (this.args.length === 4) {
                [target, effect, value, variable] = this.args;
            } else if (this.args.length === 3) {
                [target, effect, value] = this.args;
            } else {
                throw new Error('Wrong number of arguments');
            }
            if (!IDS.includes(target)) {
                throw new Error(`Invalid target "${target}", must be one of: "${IDS.join('", "')}"`);
            }
            const style = getComputedStyle(document.body);
            const h1 = parseFloat(style.getPropertyValue(`--${target}-color-h`));
            const s1 = parseFloat(style.getPropertyValue(`--${target}-color-s`).slice(0, -1));
            const l1 = parseFloat(style.getPropertyValue(`--${target}-color-l`).slice(0, -1));
            const a1 = parseFloat(style.getPropertyValue(`--${target}-color-a`));
            let [h2, s2, l2, a2] = [h1, s1, l1, a1];
            switch (effect) {
                case 'color':
                    [h2, s2, l2, a2] = RGBAtoHSLA(HEXtoRGBA(value));
                    break;
                case 'hue':
                    h2 = parseInt(value);
                    break;
                case 'shift':
                    h2 = h1 + parseInt(value);
                    break;
                case 'saturation':
                    s2 = parseFloat(value);
                    break;
                case 'saturate':
                    s2 = s1 + parseFloat(value);
                    break;
                case 'desaturate':
                    s2 = s1 - parseFloat(value);
                    break;
                case 'lightness':
                    l2 = parseFloat(value);
                    break;
                case 'lighten':
                    l2 = l1 + parseFloat(value);
                    break;
                case 'darken':
                    l2 = l1 - parseFloat(value);
                    break;
                case 'opacity':
                    a2 = parseFloat(value) / 100
                    break;
                case 'thicken':
                    a2 = a1 + parseFloat(value) / 100;
                    break;
                case 'dilute':
                    a2 = a1 - parseFloat(value) / 100;
                    break;
                default:
                    throw new Error(`Invalid ${target} effect "${effect}"`);
            }
            h2 = h2 % 360;
            s2 = Math.clamp(s2, 0, 100);
            l2 = Math.clamp(l2, 0, 100);
            a2 = Math.clamp(a2, 0, 1);
            if (isNaN(h2) || isNaN(s2) || isNaN(l2) || isNaN(a2)) {
                throw new Error(`Invalid ${effect} value "${value}"`);
            }
            output = [h2, s2, l2, a2];
            const root = document.querySelector(':root');
            root.style.setProperty(`--${target}-color-h`, h2);
            root.style.setProperty(`--${target}-color-s`, s2.toFixed(0) + '%');
            root.style.setProperty(`--${target}-color-l`, l2.toFixed(0) + '%');
            root.style.setProperty(`--${target}-color-a`, a2.toFixed(2));
            const [h3, s3, l3, _] = contrast(output);
            root.style.setProperty(`--${target}-text-h`, h3);
            root.style.setProperty(`--${target}-text-s`, s3.toFixed(0) + '%');
            root.style.setProperty(`--${target}-text-l`, l3.toFixed(0) + '%');
            if (variable) {
                if (variable[0] !== '$') {
                    throw new Error(`Invalid variable "${variable}", it should be a quoted variable name like "$accentColor"`)
                }
                State.variables[variable.slice(1)] = RGBAtoHEX(HSLAtoRGBA(output));
            }
        }
    });

    function HEXtoRGBA(value, opacity) {
        if (!value) return;
        let hex = value[0] === '#' ? value.slice(1) : value;
        const bigint = parseInt(hex, 16);
        if (hex.length == 6) {
            return [
                (bigint >> 16) & 255,
                (bigint >> 8) & 255,
                bigint & 255,
                opacity !== undefined ? parseInt(opacity) / 100 : 1.0,
            ];
        }
        if (hex.length == 8) {
            return [
                (bigint >> 24) & 255,
                (bigint >> 16) & 255,
                (bigint >> 8) & 255,
                opacity !== undefined ? parseInt(opacity) / 100 : (bigint & 255) / 255,
            ];
        }
        throw new Error(`Invalid color value "${value}", input must be a six digit hex color`)
    }

    function RGBAtoHEX(rgba) {
        const [r, g, b, a] = rgba;
        const hex = [r, g, b, Math.round(a * 255)];
        return '#' + hex.map(x => x.toString(16)).map(x => x.length == 1 ? '0' + x : x).join('').toUpperCase();
    }

    function HSLAtoRGBA(hsla) {
        let [h, s, l, a] = hsla;
        s /= 100;
        l /= 100;
        const k = n => (n + h / 30) % 12;
        const x = s * Math.min(l, 1 - l);
        const f = n => l - x * Math.max(-1, Math.min(k(n) - 3, Math.min(9 - k(n), 1)));
        return [
            Math.round(f(0) * 255),
            Math.round(f(8) * 255),
            Math.round(f(4) * 255),
            a,
        ];
    };

    function RGBAtoHSLA(rgba) {
        let [r, g, b, a] = rgba;
        r /= 255;
        g /= 255;
        b /= 255;
        // Find the minimum and maximum values of r, g, and b
        let min = Math.min(r, g, b);
        let max = Math.max(r, g, b);
        // Initialize the hue, saturation, and lightness values
        let h = 0;
        let s = 0;
        let l = (min + max) / 2;
        // If the min and max values are not the same, we need to calculate the saturation and hue values
        if (min != max) {
            let d = max - min;
            s = l > 0.5 ? d / (2 - max - min) : d / (max + min);
            if (max == r) {
                h = (g - b) / d + (g < b ? 6 : 0);
            } else if (max == g) {
                h = (b - r) / d + 2;
            } else if (max == b) {
                h = (r - g) / d + 4;
            }
            h /= 6;
        }
        return [
            Math.round(h * 360),
            Math.round(s * 100),
            Math.round(l * 100),
            a,
        ];
    }

    function contrast(hsla) {
        const [h, s, _, a] = hsla;
        const [r, g, b, __] = HSLAtoRGBA(hsla);
        const q = (r * 0.299 + g * 0.587 + b * 0.114) / 255 * 100;
        switch (true) {
            case q > 70:
                return [h, Math.min(s, 50), Math.max(q - 70, 0), a];
            case q > 50:
                return [h, 0, 0, a];
            case q > 25:
                return [h, 0, 100, a];
            default:
                return [h, Math.min(s, 50), Math.min(75 + q, 100), a];
        }
    }
}());
